# -*- coding: utf-8 -*-

# -*- coding: utf-8 -*-

import unittest

from pipesnake.transformers.financial import ToReturn


class TestFinancial(unittest.TestCase):
    def test_to_return(self):
        from _data import x
        from _data import y

        returner = ToReturn(x_cols='all', y_cols='all')
        x_new, y_new = returner.fit_transform(x, y)

        del x, y, x_new, y_new




# -*- coding: utf-8 -*-

import logging

import pandas

from pipesnake.base import Transformer
from pipesnake.base.utils import _check_cols

__all__ = [
    'Combiner',
    'Roller',
]


class Combiner(Transformer):
    """Apply user function to a column or a set of columns

    This :class:`Transformer` applies `func` to the provided list of columns or list of lists of columns.

    Args:
        :param x_cols: a list of columns name or a list of indices (or list of lists for multi-parameter func)
        :param y_cols: a list of columns name or a list of indices (or list of lists for multi-parameter func)
        :param func: the functor for the function to be applied
        :param inv_func: the functor for the inverse function to be applied
        :param as_dataframe: boolean if true the subset dataframe is passed to func (default: `False`) and the function
                             is expected to return a DataFrame
        :param reset_index: if true when the results of transformers get concatenated
                            the index are reset otherwise the original index is kept (default: `True`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> def mult(x, y):
    >>>     return x * y
    >>> combiner = Combiner(x_cols=[[2, 3], [5, 7]], func=mult)
    >>> # This will return 2 columns where each item is the combination (multiplication) of
    >>> # items taken from the provided columns
    >>> x_new, y_new = combiner.fit_transform(x, y)
    """

    def __init__(self, x_cols=[], y_cols=[], func=None, inv_func=None, as_dataframe=False, reset_index=True,
                 sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.x_cols = x_cols
        self.y_cols = y_cols
        self.func = func
        self.inv_func = inv_func
        self.as_dataframe = as_dataframe
        self.reset_index = reset_index

        # this is need to inverse mapping
        self._inverse_map_column_names = {'x': {}, 'y': {}}

    def _check_signature(self, cols):
        # check if func signature match the number of inputs
        func_signature = self.func.__code__.co_varnames
        for c in cols:
            dim = 1
            if type(c) == list:
                dim = len(c)
            assert len(func_signature) >= dim, 'mismatch number function params {} and number of inputs {}'.format(
                func_signature, c)

    def fit_x(self, x):
        self._check_signature(self.x_cols)
        existing_columns = x.columns.values.tolist()
        for cols in self.x_cols:
            if type(cols) == list:
                for c in cols:
                    assert c in existing_columns, 'column: {} not in x columns {}'.format(c, existing_columns)
            else:
                assert cols in existing_columns, 'column: {} not in x columns {}'.format(cols, existing_columns)
        return self

    def fit_y(self, y):
        self._check_signature(self.y_cols)
        existing_columns = y.columns.values.tolist()
        for cols in self.y_cols:
            if type(cols) == list:
                for c in cols:
                    assert c in existing_columns, 'column: {} not in y columns'.format(c)
            else:
                assert cols in existing_columns, 'column: {} not in y columns'.format(cols)
        return self

    def _transform(self, data, cols, proc_x=True):
        """ The actual implementation transformer

        Args:
            :param data: a pandas dataframe to be processed
            :param cols: the list of columns to be processed
            :param proc_x: if true transform `x` otherwise transform `y`
        Returns:
            :return data_new: the transformed data
        """

        # nothing to do
        if self.func is None:
            return data
        if len(cols) == 0:
            return data

        data_new = pandas.DataFrame()
        for c in cols:
            if type(c) != list:
                c = [c]
            self.logging('applying {} to {} columns'.format(self.func.__name__, c), level=logging.DEBUG)
            _x = data[c].copy()
            if type(_x) == pandas.Series:
                _x = _x.to_frame()
            if self.as_dataframe:
                data_curr = self.func(_x)
            else:
                res = []
                for _, r in _x.iterrows():
                    res.append(self.func(*r))
                data_curr = pandas.DataFrame(res)

            curr_column_name = self.func.__name__ + '_' + '_'.join([str(i) for i in c])
            data_curr.columns = [curr_column_name]

            if proc_x:
                self._inverse_map_column_names['x'][curr_column_name] = c
            else:
                self._inverse_map_column_names['y'][curr_column_name] = c

            # concatenate columns
            join = True
            if data_new.empty:
                data_new = data_curr.copy()
                if self.reset_index:
                    data_new.reset_index(inplace=True)
                data_new.drop(['index'], axis=1, inplace=True)
                join = False
            if join:
                if self.reset_index:
                    data_curr.reset_index(inplace=True)
                data_curr.drop(['index'], axis=1, inplace=True)
                data_new = data_new.join(data_curr)

        return data_new

    def transform_x(self, x):
        return self._transform(x, self.x_cols)

    def transform_y(self, y):
        return self._transform(y, self.y_cols, proc_x=False)

    def _inverse_transform(self, data, proc_x=True):
        """ The actual implementation of inverse transform

        Args:
            :param data:
            :param proc_x:
        """

        # nothing to do
        if self.inv_func is None:
            return data
        if proc_x and len(self._inverse_map_column_names['x']) == 0:
            return data
        if not proc_x and len(self._inverse_map_column_names['y']) == 0:
            return data

        if proc_x:
            cols = self._inverse_map_column_names['x']
        else:
            cols = self._inverse_map_column_names['y']

        data_org = pandas.DataFrame()
        for col, org_cols in cols.items():
            self.logging('inverting from {} to {} columns using {}'.format(col, org_cols, self.inv_func.__name__),
                         level=logging.DEBUG)
            _t = data[col].copy().to_frame()
            if type(_t) == pandas.Series:
                _t = _t.to_frame()
            if self.as_dataframe:
                data_curr = self.inv_func(_t)
            else:
                res = []
                for _, r in _t.iterrows():
                    res.append(self.inv_func(*r))
                data_curr = pandas.DataFrame(res)
            if type(org_cols) != list:
                org_cols = [org_cols]
            data_curr.columns = org_cols
            logging.debug('expected cols: {}'.format(org_cols))
            logging.debug('current cols: {}'.format(data_curr.columns.values.tolist()))
            assert len(org_cols) == len(
                data_curr.columns.values.tolist()), 'mismatch number of cols {} expected {}'.format(
                len(org_cols), len(data_curr.columns.values.tolist()))

            # concatenate columns
            join = True
            if data_org.empty:
                data_org = data_curr.copy()
                if self.reset_index:
                    data_org.reset_index(inplace=True)
                data_org.drop(['index'], axis=1, inplace=True)
                join = False
            if join:
                if self.reset_index:
                    data_curr.reset_index(inplace=True)
                data_curr.drop(['index'], axis=1, inplace=True)
                data_org = data_org.join(data_curr)

        return data_org

    def inverse_transform_x(self, x):
        return self._inverse_transform(x)

    def inverse_transform_y(self, y):
        return self._inverse_transform(y, proc_x=False)


class Roller(Combiner):
    """Apply the provided function rolling within a given window

    Note: resulting columns may contain nan (typically a total of `window-1` at the beginning)

    Args:
        :param x_cols: a list of columns name or a list of indices (or list of lists for multi-parameter func)
        :param y_cols: a list of columns name or a list of indices (or list of lists for multi-parameter func)
        :param func: the functor for the function to be applied (default: `None`)
        :param inv_func: the functor for the inverse function to be applied (default: `None`)
        :param window: an integer indicating the window size (default: `3`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> roller = Roller(x_cols=[0], roll_func=lambda x: x[0] + x[1], window=2)
    >>> x_new, y_new = roller.fit_transform(x, y)
    """

    def __init__(self, x_cols=[], y_cols=[], roll_func=None, inv_roll_func=None, window=3, sklearn_output=False,
                 name=None):
        Combiner.__init__(self, x_cols=x_cols, y_cols=y_cols, func=self.roller_func, inv_func=self.roller_inv_func,
                          as_dataframe=True, sklearn_output=sklearn_output, name=name)
        self.roll_func = roll_func
        self.inv_roll_func = inv_roll_func
        self.window = window

    def roller_func(self, data):
        return data.rolling(window=self.window).apply(self.roll_func)

    def roller_inv_func(self, data):
        return data.rolling(window=self.window).apply(self.inv_roll_func)

    def fit_x(self, x):
        Combiner._check_signature(self, self.x_cols)
        self.x_cols = _check_cols(x, self.x_cols, self.logging)
        return self

    def fit_y(self, y):
        Combiner._check_signature(self, self.y_cols)
        self.y_cols = _check_cols(y, self.y_cols, self.logging)
        return self



# -*- coding: utf-8 -*-
import logging

import pandas

from pipesnake.base import Transformer
from pipesnake.base.utils import _check_cols

__all__ = [
    'Category2Number',
]


class Category2Number(Transformer):
    """Convert categorical to number.

    Args:
        :param x_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param y_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param skipna: if True the null values (None, NaN, ...) will not be replaced (default: `False`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> cat2num = Category2Number(x_cols=['x_1'], y_cols=['y_1'])
    >>> x_new, y_new = cat2num.fit_transform(x, y)
    """

    def __init__(self, x_cols=[], y_cols=[], skipna=False, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.x_cols = x_cols
        self.y_cols = y_cols
        self.skipna = skipna
        self._map_category_to_number = {'x': {}, 'y': {}}
        self._map_number_to_category = {'x': {}, 'y': {}}

    def _fit(self, data, cols):
        cat_to_num = {}
        num_to_cat = {}
        for c in cols:
            cat_to_num[c] = {}
            num_to_cat[c] = {}
            uniques = sorted(data[c].unique().tolist())
            for i, v in enumerate(uniques):
                if pandas.isnull(v):
                    if self.skipna:
                        continue
                    else:
                        cat_to_num[c]['_nan_'] = i
                        num_to_cat[c][i] = pandas.np.nan
                else:
                    cat_to_num[c][v] = i
                    num_to_cat[c][i] = v
        return cat_to_num, num_to_cat

    def fit_x(self, x):
        """Create the list of columns to be converted for `x` and related mappings.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Return:
            :param self:
        """
        self.x_cols = _check_cols(x, self.x_cols, self.logging)
        self.logging('x_cols: {}'.format(self.x_cols), level=logging.DEBUG)
        self._map_category_to_number['x'], self._map_number_to_category['x'] = self._fit(x, self.x_cols)
        return self

    def fit_y(self, y):
        """Create the list of columns to be converted for `y` and related mappings.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Return:
            :param self:
        """
        self.y_cols = _check_cols(y, self.y_cols, self.logging)
        self.logging('y_cols: {}'.format(self.y_cols), level=logging.DEBUG)
        self._map_category_to_number['y'], self._map_number_to_category['y'] = self._fit(y, self.y_cols)
        return self

    def _transform(self, data, cols, cat_to_num):
        if len(cols) == 0:
            return data
        _data = data.copy()
        if not self.skipna:
            _data[cols].fillna(value='_nan_', inplace=True)
        for c in cols:
            _data.replace(to_replace={c: cat_to_num[c]}, inplace=True)
        return _data

    def transform_x(self, x):
        self.logging('x category to number: {}'.format(self._map_category_to_number['x']), level=logging.DEBUG)
        return self._transform(x, self.x_cols, self._map_category_to_number['x'])

    def transform_y(self, y):
        self.logging('y category to number: {}'.format(self._map_category_to_number['y']), level=logging.DEBUG)
        return self._transform(y, self.y_cols, self._map_category_to_number['y'])

    def _inverse_transform(self, data, cols, num_to_cat):
        if len(cols) == 0:
            return data
        _data = data.copy()
        for c in cols:
            _data.replace(to_replace={c: num_to_cat[c]}, inplace=True)
        return _data

    def inverse_transform_x(self, x):
        self.logging('x number to category: {}'.format(self._map_number_to_category['x']), level=logging.DEBUG)
        return self._inverse_transform(x, self.x_cols, self._map_number_to_category['x'])

    def inverse_transform_y(self, y):
        self.logging('y number to category: {}'.format(self._map_number_to_category['y']), level=logging.DEBUG)
        return self._inverse_transform(y, self.y_cols, self._map_number_to_category['y'])



# -*- coding: utf-8 -*-

import logging

import pandas

from pipesnake.base import Pipe
from pipesnake.base import Transformer
from pipesnake.utils.decorator import memory
from pipesnake.utils.decorator import timeit

__all__ = [
    'ParallelPipe',
]


class ParallelPipe(Pipe, Transformer):
    """:class:`ParallelPipe` applies a list of transformers as parallel.

    The class itself inherits from :class:`Pipe` and :class:`Transformer` so that it is possible
    to use this as :class:`Transformer` objects.

    Args:
        :param transformers: a list of :class:`Transformer` objects
        :param reset_index: if true when the results of transformers get concatenated
                                the index are reset otherwise the original index is kept
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline
        :param name: name for this :class:`Transformer`
    """

    def __init__(self, transformers=[], reset_index=True, sklearn_output=False, name=None):
        Pipe.__init__(self, transformers=transformers)
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.reset_index = reset_index
        self._original_column_names = {'x': {}, 'y': {}}

    @memory
    @timeit
    def fit_x(self, x):
        """Fit all transformers parameters in the pipe.

        This function apply `fit_x` for each transformer in the pipe.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:
            :return self:

        .. todo:: This can be parallelized in computation
        """
        self.logging('fitting x...')
        self._original_column_names['x'] = x.columns
        for t in self.transformers:
            self.logging('-> {}'.format(t.name))
            t.fit_x(x)
        return self

    @memory
    @timeit
    def fit_y(self, y):
        """Fit all transformers parameters in the pipe.

        This function apply `fit_y` for each transformer in the pipe.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return self:

        .. todo:: This can be parallelized in computation
        """
        self.logging('fitting y...')
        self._original_column_names['y'] = y.columns
        for t in self.transformers:
            self.logging('-> {}'.format(t.name))
            t.fit_y(y)
        return self

    def _transform(self, data, proc_x=True):
        """ The actual implementation transformer

        Args:
            :param data: a pandas dataframe
            :param proc_x: if true transform `x` otherwise transform `y`
        Returns:
            :return data_new: the transformed data

        .. todo:: This can be parallelized in computation
        """

        # nothing to do
        if len(self.transformers) == 0:
            return data

        data_new = pandas.DataFrame()
        for t in self.transformers:
            self.logging('-> {}'.format(t.name))
            if proc_x:
                data_curr = t.transform_x(data)
            else:
                data_curr = t.transform_y(data)
            if data_curr is None or data_curr.empty:
                continue

            join = True
            if data_new.empty:
                data_new = data_curr.copy()
                if self.reset_index:
                    data_new.reset_index(inplace=True)
                data_new.drop(['index'], axis=1, inplace=True)
                data_new.columns = [c + '_{}'.format(t.name) for c in data_new.columns]
                join = False

            if join:
                if self.reset_index:
                    data_curr.reset_index(inplace=True)
                data_curr.drop(['index'], axis=1, inplace=True)
                data_curr.columns = [c + '_{}'.format(t.name) for c in data_curr.columns]
                data_new = data_new.join(data_curr)

        return data_new

    @memory
    @timeit
    def transform_x(self, x):
        """Transform `x`.

        This function apply parallel `transform` for each transformer in the pipe concatenating
        the results.

        As far as the input for each transformer is the same dataset after the tranformations
        the return columns are renamed like: `<original_col_name>_<transformer_name>`.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:
            :return x_new: the new transformed x
        """
        self.logging('transforming x...')
        return self._transform(x)

    @memory
    @timeit
    def transform_y(self, y):
        """Transform `y`.

        This function apply parallel `transform` for each transformer in the pipe concatenating
        the results.

        As far as the input for each transformer is the same dataset after the tranformations
        the return columns are renamed like: `<original_col_name>_<transformer_name>`.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return y_new: the new transformed y
        """
        self.logging('transforming y...')
        return self._transform(y, proc_x=False)

    @memory
    @timeit
    def inverse_transform_x(self, x):
        """Inverse transform `x`.

        As far this applies transformers in parallel is good enough to invert just one of them
        to get back to the original inputs.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:
            :return x_org: the original inverse transformed x
        """
        self.logging('inverse transforming x...')
        x_org = None
        for t in reversed(self.transformers):
            self.logging('-> {}'.format(t.name))
            x_r = t.inverse_transform_x(x)
            if x_r is not None:
                x_org = x_r.copy()
            if x_org is not None:
                break
        assert x_org is not None, 'has been impossible to inverse_transform x'
        self.logging('original x columns: {}'.format(self._original_column_names['x']), level=logging.DEBUG)
        self.logging('inverted x columns: {}'.format(x_org.columns), level=logging.DEBUG)
        assert len(x_org.columns) == len(
            self._original_column_names['x']), 'mismatch number of cols {} expected {}'.format(len(x_org.columns), len(
            self._original_column_names['x']))
        return x_org

    @memory
    @timeit
    def inverse_transform_y(self, y):
        """Inverse transform `y`.

        As far this applies transformers in parallel is good enough to invert just one of them
        to get back to the original inputs.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return y_org: the original inverse transformed y
        """
        self.logging('inverse transforming y...')
        y_org = None
        for t in reversed(self.transformers):
            self.logging('-> {}'.format(t.name))
            y_r = t.inverse_transform_y(y)
            if y_r is not None:
                y_org = y_r.copy()
            if y_org is not None:
                break
        assert y_org is not None, 'has been impossible to inverse_transform y'
        self.logging('original y columns: {}'.format(self._original_column_names['y']), level=logging.DEBUG)
        self.logging('inverted y columns: {}'.format(y_org.columns), level=logging.DEBUG)
        assert len(y_org.columns) == len(
            self._original_column_names['y']), 'mismatch number of cols {} expected {}'.format(len(y_org.columns), len(
            self._original_column_names['y']))
        return y_org



# -*- coding: utf-8 -*-

"""
The :mod:`pipesnake.transformers` module implements several transformers.

Here is a short list of the main modules:

:mod:`pipesnake.transformers.combiner`
:mod:`pipesnake.transformers.converter`
:mod:`pipesnake.transformers.deeplearning`
:mod:`pipesnake.transformers.dropper`
:mod:`pipesnake.transformers.financial`
:mod:`pipesnake.transformers.imputer`
:mod:`pipesnake.transformers.misc`
:mod:`pipesnake.transformers.scaler`
:mod:`pipesnake.transformers.selector`
:mod:`pipesnake.transformers.stats`
"""



# -*- coding: utf-8 -*-

import unittest

import pandas

from pipesnake.pipe import SeriesPipe
from pipesnake.transformers.imputer import KnnImputer
from pipesnake.transformers.imputer import ReplaceImputer
from pipesnake.transformers.selector import ColumnSelector


class TestImputer(unittest.TestCase):
    def test_replace_imputer(self):
        from _data import x_nan
        from _data import y_nan

        pipe = SeriesPipe(transformers=[
            ColumnSelector(x_cols=['x_1', 'x_2', 'x_3', 'x_4'], y_cols=['y_1', 'y_2']),
            ReplaceImputer(x_cols='all', y_cols='all'),
        ])
        x_new, y_new = pipe.fit_transform(x_nan, y_nan)
        self.assertEqual(pandas.isnull(x_new).any().any(), False, 'NaN values has been found in x_new')
        self.assertEqual(pandas.isnull(y_new).any().any(), False, 'NaN values has been found in y_new')

        del x_nan, y_nan, x_new, y_new

    def test_knn_imputer(self):
        from _data import x_nan
        from _data import y_nan

        imputer = KnnImputer(x_cols='all', y_cols='all')
        x_new, y_new = imputer.fit_transform(x_nan, y_nan)
        self.assertEqual(pandas.isnull(x_new).any().any(), False, 'NaN values has been found in x_new')
        self.assertEqual(pandas.isnull(y_new).any().any(), False, 'NaN values has been found in y_new')

        del x_nan, y_nan, x_new, y_new



# -*- coding: utf-8 -*-

import logging
import os

from pylint import lint


def get_py_files(base_path):
    f = []
    for root, dirs, files in os.walk(base_path):
        for name in files:
            if name.endswith('.py'):
                f.append(os.path.normpath(os.path.join(root, name)))
    return f


def main():
    params = ['-j', '4', '--msg-template={path}:{line}: [{msg_id}({symbol}), {obj}] {msg}']
    files = get_py_files('pipesnake')
    params.extend(files)
    logging.info('pylint params: %s', ' '.join(params))
    logging.info('Going to lint: %s files', len(files))
    lint.Run(params)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

import unittest

from pipesnake.transformers.misc import ColumnRenamer


class TestBase(unittest.TestCase):
    def test_renamer(self):
        from _data import x
        from _data import y

        renamer = ColumnRenamer(prefix='test_')
        x_new, y_new = renamer.fit_transform(x, y)

        for c in x_new.columns.values:
            self.assertEqual(c.startswith('test_x_'), True, 'wrong column name: {}'.format(c))

        for c in y_new.columns.values:
            self.assertEqual(c.startswith('test_y_'), True, 'wrong column name: {}'.format(c))

        x_org, y_org = renamer.inverse_transform(x_new, y_new)

        for c in x_org.columns.values:
            self.assertEqual(c.startswith('x_'), True, 'wrong column name: {}'.format(c))

        for c in y_org.columns.values:
            self.assertEqual(c.startswith('y_'), True, 'wrong column name: {}'.format(c))

        del x, y, x_new, y_new, x_org, y_org



# -*- coding: utf-8 -*-


import unittest

from pipesnake.pipe import ParallelPipe
from pipesnake.pipe import SeriesPipe
from pipesnake.transformers.selector import ColumnSelector
from pipesnake.utils.graph import to_graph


class TestGraph(unittest.TestCase):
    def test_graph(self):
        pipe = SeriesPipe(transformers=[
            ParallelPipe(transformers=[
                ColumnSelector(x_cols=['x_1', 'x_2', 'x_3', 'x_4', ], y_cols=['y_1', ], name='1'),
                ColumnSelector(x_cols=['x_3', ]),
            ]),
            ColumnSelector(x_cols=['x_1_1', ], y_cols=['y_1_1', ]),
        ])
        graph = to_graph(pipe)
        for node in graph:
            print(node, graph.node[node]['params'])



# -*- coding: utf-8 -*-

import logging

import joblib

__all__ = [
    'dump_pipe',
    'load_pipe',
]


def dump_pipe(pipe, filename, compress=True):
    """Dump this pipe on file using JobLib.

    :param pipe: the Pipe inherited object to dump on file
    :param filename: the filename to use
    :param compress: is True use compressed file
    """
    logging.info('dumping: {} on {}'.format(pipe.name, filename))
    with open(filename, 'wb') as _file:
        joblib.dump(pipe, _file, compress=compress)


def load_pipe(filename):
    """Load pipe from file using JobLib.

    :param filename: the filename to load
    """
    logging.info('loading: {}'.format(filename))
    with open(filename, 'rb') as _file:
        return joblib.load(_file)



# -*- coding: utf-8 -*-

import unittest

from pipesnake.transformers.combiner import Combiner
from pipesnake.transformers.combiner import Roller


class TestCombiner(unittest.TestCase):
    def test_combiner(self):
        from _data import x
        from _data import y

        def foo(x, y):
            return x + y + 1.0

        def inv_foo(x):
            t = (x - 1.0) / 2.0
            return t, t

        combiner = Combiner(x_cols=[['x_1', 'x_4'], ['x_2', 'x_3']],
                            y_cols=[['y_1', 'y_2']],
                            func=foo,
                            inv_func=inv_foo)
        x_new, y_new = combiner.fit_transform(x, y)
        self.assertEqual(x_new.shape[1], 2, 'mismatch sizes: {} != 2'.format(x_new.shape[1]))
        self.assertEqual(y_new.shape[1], 1, 'mismatch sizes: {} != 1'.format(y_new.shape[1]))

        for _, r in x_new.iterrows():
            self.assertEqual(r[0], r[1], 'mismatch size: {} != {}'.format(r[0], r[1]))

        for l, r, in zip(y.values.tolist(), y_new.values.tolist()):
            self.assertEqual(foo(l[0], l[1]), r[0], 'mismatch value: {} != {}'.format(l, r))

        x_org, y_org = combiner.inverse_transform(x_new, y_new)
        self.assertEqual(x_org.shape[1], 4, 'mismatch sizes: {} != 4'.format(x_org.shape[1]))
        self.assertEqual(y_org.shape[1], 2, 'mismatch sizes: {} != 2'.format(y_org.shape[1]))

        for r in y_org.values.tolist():
            self.assertEqual(r[0], r[1], 'mismatch value: {} != {}'.format(r[0], r[1]))

        del x, y, x_new, y_new, x_org, y_org

    def test_roller(self):
        from _data import x
        from _data import y

        roller = Roller(x_cols='all', y_cols='all', roll_func=lambda x: x[1], inv_roll_func=lambda x: x[1], window=2)
        x_new, y_new = roller.fit_transform(x, y)
        x_org, y_org = roller.inverse_transform(x_new, y_new)

        x_new.columns = x.columns  # in order to de able to use == below
        y_new.columns = y.columns  # in order to de able to use == below

        equals = (x == x_new).any(1)
        for i, (left, right) in enumerate(zip([False, True, True, True], equals)):
            self.assertEqual(left, right, 'mismatched values in x row {}: {} != {}'.format(i, left, right))
        equals = (y == y_new).any(1)
        for i, (left, right) in enumerate(zip([False, True, True, True], equals)):
            self.assertEqual(left, right, 'mismatched values in y row {}: {} != {}'.format(i, left, right))

        equals = (x == x_org).any(1)
        for i, (left, right) in enumerate(zip([False, False, True, True], equals)):
            self.assertEqual(left, right, 'mismatched values in x row {}: {} != {}'.format(i, left, right))
        equals = (y == y_org).any(1)
        for i, (left, right) in enumerate(zip([False, False, True, True], equals)):
            self.assertEqual(left, right, 'mismatched values in y row {}: {} != {}'.format(i, left, right))

        del x, y, x_new, y_new, x_org, y_org



# -*- coding: utf-8 -*-

"""
The :mod:`pipesnake` is yet another pandas sklearn-inspired pipeline data processor.
"""

# Set default logging handler to avoid "No handler found" warnings.
import logging

from pipesnake import metadata

__version__ = metadata.version
__author__ = metadata.authors[0]
__license__ = metadata.license
__copyright__ = metadata.copyright

try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())



# -*- coding: utf-8 -*-

from pipesnake.base import Pipe
from pipesnake.base import Transformer
from pipesnake.utils.decorator import memory
from pipesnake.utils.decorator import timeit

__all__ = [
    'SeriesPipe',
]


class SeriesPipe(Pipe, Transformer):
    """:class:`SeriesPipe` applies a list of transformers as series.

    The class itself inherits from :class:`Pipe` and :class:`Transformer` so that it is possible
    to use this as :class:`Transformer` objects.

    Args:
        :param transformers: a list of :class:`Transformer` objects
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline
        :param name: name for this :class:`Transformer`
    """

    def __init__(self, transformers=[], sklearn_output=False, name=None):
        Pipe.__init__(self, transformers=transformers)
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)

    @memory
    @timeit
    def fit_x(self, x):
        """Fit all transformers parameters in the pipe on `x`.

        This function apply sequentially `fit` for each transformer in the pipe.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:
            :return self:

        As far as the `fit` of subsequent transformers may be affected by the previous one
        is necessary to execute `fit_transform` here.

        .. todo:: Add some cache system in order to do not recompute this stuff in `transform`.
        """
        self.logging('fitting x...')
        x_new = x.copy()
        for t in self.transformers:
            self.logging('-> {}'.format(t.name))
            x_new = t.fit_transform_x(x_new)
        return self

    @memory
    @timeit
    def fit_y(self, y):
        """Fit all transformers parameters in the pipe on `y`.

        This function apply sequentially `fit` for each transformer in the pipe.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return self:

        As far as the `fit` of subsequent transformers may be affected by the previous one
        is necessary to execute `fit_transform` here.

        .. todo:: Add some cache system in order to do not recompute this stuff in `transform`.
        """
        self.logging('fitting y...')
        y_new = y.copy()
        for t in self.transformers:
            self.logging('-> {}'.format(t.name))
            y_new = t.fit_transform_y(y_new)
        return self

    @memory
    @timeit
    def transform_x(self, x):
        """Transform `x`.

        This function apply sequentially `transform` for each transformer in the pipe.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:
            :return x_new: the new transformed x
        """
        self.logging('transforming x...')
        x_new = x.copy()
        for t in self.transformers:
            self.logging('-> {}'.format(t.name))
            x_new = t.transform_x(x_new)
        return x_new

    @memory
    @timeit
    def transform_y(self, y):
        """Transform `y`.

        This function apply sequentially `transform` for each transformer in the pipe.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return y_new: the new transformed y
        """
        self.logging('transforming y...')
        y_new = y.copy()
        for t in self.transformers:
            self.logging('-> {}'.format(t.name))
            y_new = t.transform_y(y_new)
        return y_new

    @memory
    @timeit
    def inverse_transform_x(self, x):
        """Inverse transform `x`.

        This function apply `inverse_transform` for each transformer in the pipe to get back to the original x and y.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:
            :return x_new: the new inverse transformed x
        """
        self.logging('inverse transforming x...')
        x_new = x.copy()
        for t in reversed(self.transformers):
            self.logging('-> {}'.format(t.name))
            x_new = t.inverse_transform_x(x_new)
        return x_new

    @memory
    @timeit
    def inverse_transform_y(self, y):
        """Inverse transform `y`.

        This function apply `inverse_transform` for each transformer in the pipe to get back to the original x and y.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return y_new: the new inverse transformed y
        """
        self.logging('inverse transforming y...')
        y_new = y.copy()
        for t in reversed(self.transformers):
            self.logging('-> {}'.format(t.name))
            y_new = t.inverse_transform_y(y_new)
        return y_new



# -*- coding: utf-8 -*-

import logging

import numpy

from pipesnake.base import Transformer
from pipesnake.base.utils import _check_cols
from pipesnake.utils.decorator import memoize

__all__ = [
    'ToSymbolProbability',
]


@memoize
def get_pdf_value(hist, edges, pos):
    """Get the value of a pdf for a given x

    :param hist: list of histogram values
    :param edges: list of bins
    :param pos: position on the support
    :return: pdf value
    """
    ed = numpy.digitize(pos, edges).tolist()
    if ed > len(hist) - 1:
        return hist[-1]
    return hist[ed]


@memoize
def get_pdf_support(edges, pos):
    """As far as the support is quantized this maps a
    continuous value to a quantized one

    :param edges: list of bins
    :param pos: position on the support
    :return: quantized position
    """
    return edges[numpy.digitize(pos, edges).tolist()]


def build_support(x):
    """Build the support for the x pdf."""
    _x = numpy.sort(numpy.unique(x))
    _delta = numpy.median(numpy.diff(_x))
    _min = numpy.min(_x)
    _max = numpy.max(_x)
    return numpy.arange(_min - _delta, _max + _delta, _delta)


def build_pdf(x, density=True):
    """Compute the histogram (pdf) of the data"""
    hist, edges = numpy.histogram(x, bins=build_support(x), density=density)
    return hist, edges


class ToSymbolProbability(Transformer):
    """Convert values in columns to their probabilities.

    The transfomer estimate the histogram (`pdf`) for the provided columns and
    then convert original values to probability `x_new = pdf(x)`

    Args:
        :param x_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param y_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> topb = ToSymbolProbability(x_cols=['x_1'])
    >>> x_new, y_new = topb.fit_transform(x, y)
    """

    def __init__(self, x_cols=[], y_cols=[], sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.x_cols = x_cols
        self.y_cols = y_cols
        self._pdfs = {'x': {}, 'y': {}}

    def _fit(self, data, cols, pdf):
        for c in cols:
            self.logging(' {}'.format(c), level=logging.DEBUG)
            pdf[c] = build_pdf(data[c])

    def fit_x(self, x):
        self.x_cols = _check_cols(x, self.x_cols, self.logging)
        self.logging('x_cols: {}'.format(self.x_cols), level=logging.DEBUG)
        self.logging('fitting x', level=logging.DEBUG)
        self._fit(x, self.x_cols, self._pdfs['x'])
        return self

    def fit_y(self, y):
        self.y_cols = _check_cols(y, self.y_cols, self.logging)
        self.logging('y_cols: {}'.format(self.y_cols), level=logging.DEBUG)
        self.logging('fitting y', level=logging.DEBUG)
        self._fit(y, self.y_cols, self._pdfs['y'])
        return self

    def _transform(self, data, cols, pdf):
        _data = data.copy()
        for c in cols:
            self.logging(' {}'.format(c), level=logging.DEBUG)
            _data[c] = _data[c].apply(lambda x: get_pdf_value(pdf[c][0], pdf[c][1], x))
        return _data

    def transform_x(self, x):
        self.logging('transforming x', level=logging.DEBUG)
        return self._transform(x, self.x_cols, self._pdfs['x'])

    def transform_y(self, y):
        self.logging('transforming x', level=logging.DEBUG)
        return self._transform(y, self.y_cols, self._pdfs['y'])

    def inverse_transform_x(self, x):
        self.logging('ToSymbolProbability on x is not invertible', level=logging.WARNING)
        return x

    def inverse_transform_y(self, y):
        self.logging('ToSymbolProbability on y is not invertible', level=logging.WARNING)
        return y



# -*- coding: utf-8 -*-

import functools
import logging
import os
import time
from hashlib import sha256

import psutil

__all__ = [
    'timeit',
    'memory',
    'memoize',
]


def timeit(method):
    """ A decorator to measure the time spent by a function.
    """

    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        logging.debug('Function: {}: {:.2f} sec'.format(method.__name__, te - ts))
        return result

    return timed


def memory(method):
    """ A decorator to measure the memory before and after the function
    """

    def memory_usage_psutil(*args, **kw):
        process = psutil.Process(os.getpid())
        logging.debug(
            'Function: {} before Memory: {:.2f} MB'.format(method.__name__, process.memory_info()[0] / float(2 ** 20)))
        result = method(*args, **kw)
        logging.debug(
            'Function: {} after Memory: {:.2f} MB'.format(method.__name__, process.memory_info()[0] / float(2 ** 20)))
        return result

    return memory_usage_psutil


def memoize(obj):
    """Memoizer"""
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = sha256((str(args) + str(kwargs)).encode('utf-8'))
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]

    return memoizer



# -*- coding: utf-8 -*-

import logging
from abc import ABCMeta
from abc import abstractmethod
from copy import deepcopy
from random import getrandbits

from sklearn.base import BaseEstimator

from pipesnake.base.utils import _check_input
from pipesnake.base.utils import _check_transformer_type
from pipesnake.utils import to_snake

__all__ = [
    'Transformer',
    'Pipe',
]


class Transformer(BaseEstimator):
    """Transformer abstract class freely inspired by Scikit-Learn.

     See `TransformerMixin <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/base.py>`_
     for more details, the main differences are the parameters type and the returns.
     Note: that this inherits from sklearn BaseEstimator to simplify the parameters managment.
     """
    __metaclass__ = ABCMeta

    def __init__(self, name=None, sklearn_output=False, **kwargs):
        """
        :param name: provide a name for this transfomer
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline
        """
        self.name = name
        if name is None:
            # if no name is provided name will be assigned as the class name with a random hex
            self.name = to_snake(self.__class__.__name__) + '_{:02x}'.format(getrandbits(16))
        self.sklearn_output = sklearn_output

    @abstractmethod
    def fit_x(self, x):
        """Fit the Transformer parameters for `x`.

        This function is intended for implementing the
        computation of the parameters needed for transform function.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:
            :return self:
        Raises:
            `NotImplementedError`: function not implemented
        """
        raise NotImplementedError()

    @abstractmethod
    def fit_y(self, y):
        """Fit the Transformer parameters for `y`.

        This function is intended for implementing the
        computation of the parameters needed for transform function.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return self:
        Raises:
            `NotImplementedError`: function not implemented
        """
        raise NotImplementedError()

    def fit(self, x, y=None):
        """Fit the Transformer parameters.

        If `x` or `y` are not Pandas DataFrame not fit will be performed

        Note: `y=None` as default is to be compliant with scikit-learn

        Returns:
            :return self:
        """
        self.logging('fitting...', level=logging.DEBUG)
        if _check_input(x, self.logging):
            self.fit_x(x)
        if _check_input(y, self.logging):
            self.fit_y(y)
        return self

    @abstractmethod
    def transform_x(self, x):
        """Transform `x`.

        This function is intended to implement the actual transformation on
        x.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:
            :return x_new: the new transformed x
        Raises:
            `NotImplementedError`: function not implemented
        """
        raise NotImplementedError()

    @abstractmethod
    def transform_y(self, y):
        """Transform `y`.

        This function is intended to implement the actual transformation on
        y.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return y_new: the new transformed y
        Raises:
            `NotImplementedError`: function not implemented
        """
        raise NotImplementedError()

    def transform(self, x, y=None):
        """Transform `x` and `y`.

        If `x` or `y` are not Pandas DataFrame the original value will be returned

        Note: `y=None` as default is to be compliant with scikit-learn

        Returns:
            :return x_new: the new transformed x
            :return y_new: the new transformed y

        .. todo:: This can be parallelized in computation on x and y
        """
        self.logging('transforming...', level=logging.DEBUG)
        _x = x
        if _check_input(x, self.logging):
            _x = self.transform_x(x)
        if self.sklearn_output:
            return _x
        _y = y
        if _check_input(y, self.logging):
            _y = self.transform_y(y)
        return _x, _y

    def fit_transform(self, x, y=None):
        """Apply `fit` and `transform` functions.

        Note: `y=None` as default is to be compliant with scikit-learn

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return x_new: the new transformed x
            :return y_new: the new transformed y
        """
        return self.fit(x, y).transform(x, y)

    def fit_transform_x(self, x):
        """Apply `fit` and `transform` functions only on `x`.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:
            :return x_new: the new transformed x
        """
        return self.fit_x(x).transform_x(x)

    def fit_transform_y(self, y):
        """Apply `fit` and `transform` functions only on `y`.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return y_new: the new transformed y
        """
        return self.fit_y(y).transform_y(y)

    @abstractmethod
    def inverse_transform_x(self, x):
        """Inverse transform `x`.

        This function is intended to implement the inverse
        transform to get back to the original x.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:
            :return x_org: the original inverse transformed x
        Raises:
            `NotImplementedError`: function not implemented
        """
        raise NotImplementedError()

    @abstractmethod
    def inverse_transform_y(self, y):
        """Inverse transform `y`.

        This function is intended to implement the inverse
        transform to get back to the original y.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return y_org: the original inverse transformed y
        Raises:
            `NotImplementedError`: function not implemented
        """
        raise NotImplementedError()

    def inverse_transform(self, x, y=None):
        """Inverse transform `x` and `y`.

        This function is intended to implement the inverse
        transform to get back to the original x and y.

        Note: `y=None` as default is to be compliant with scikit-learn

        Returns:
            :return x_org: the new inverse transformed x
            :return y_org: the new inverse transformed y
        """
        self.logging('inverse transforming...', level=logging.DEBUG)
        _x = x
        if _check_input(x, self.logging):
            _x = self.inverse_transform_x(x)
        if self.sklearn_output:
            return _x
        _y = y
        if _check_input(y, self.logging):
            _y = self.inverse_transform_y(y)
        return _x, _y

    def logging(self, msg, level=logging.INFO):
        """Helper function to log info related to Transformer

        Args:
            :param msg: the message to log
            :param level: logging level enum (https://docs.python.org/2/library/logging.html#logging-levels)
        """
        logging.log(level, '[{}] : {}'.format(self.name, msg))


class Pipe(object):
    """Pipe abstract class to apply a list of transformers.

    This provides all basic wiring to deal with list of transformers.

    :param name: a name for this pipe object
    :param transformers: a list of transfomers
    """

    def __init__(self, transformers=[]):
        self.transformers = transformers

    def __len__(self):
        return len(self.transformers)

    def __iter__(self):
        return iter(self.transformers)

    def __str__(self):
        s = '['
        for c in self.transformers:
            s += str(c) + ', '
        return s[:-1] + ']'

    def __getitem__(self, index):
        return self.transformers[index]

    def __setitem__(self, index, transformer):
        assert _check_transformer_type(transformer, self.logging), 'Mismatched type: expected Transformer got {0}'.format(
            type(transformer))
        self.transformers[index] = transformer

    def __delitem__(self, index):
        del self.transformers[index]

    def copy(self, transformers):
        """Copy a list of :class:`Transformer` objects to this object."""
        del self.transformers[:]
        for t in transformers:
            self.transformers.append(t)

    def clone(self):
        """Clone this object."""
        return deepcopy(self)

    def append(self, transformer):
        assert _check_transformer_type(transformer, self.logging), 'Mismatched type: expected Transformer got {0}'.format(
            type(transformer))
        self.transformers.append(transformer)

    def extend(self, transformers):
        for t in transformers:
            self.append(t)



# -*- coding: utf-8 -*-

import unittest

from numpy import int64

from pipesnake.transformers.converter import Category2Number


class TestConverter(unittest.TestCase):
    def test_category_to_number(self):
        from _data import x_str
        from _data import y_str

        cat2num = Category2Number(x_cols=['x_5'], y_cols=['y_3'])
        x_new, y_new = cat2num.fit_transform(x_str, y_str)

        for item in x_new['x_5'].values:
            self.assertEqual(type(item) == int64, True, 'error: {} should be an int got: {}'.format(item, type(item)))

        for item in y_new['y_3'].values:
            self.assertEqual(type(item) == int64, True, 'error: {} should be an int got: {}'.format(item, type(item)))

        x_org, y_org = cat2num.inverse_transform(x_new, y_new)

        for a, b in zip(x_str['x_5'].values, x_org['x_5'].values):
            self.assertEqual(a, b, 'mismatch value {} != {}'.format(a, b))

        for a, b in zip(y_str['y_3'].values, y_org['y_3'].values):
            self.assertEqual(a, b, 'mismatch value {} != {}'.format(a, b))

        del x_str, y_str, x_new, y_new



# -*- coding: utf-8 -*-

import unittest

from pipesnake.pipe import SeriesPipe


class TestBase(unittest.TestCase):
    def test(self):
        from _transformer import FakeTransformer

        transformers = [FakeTransformer('fake1'), FakeTransformer('fake2'), ]
        series = SeriesPipe(transformers=transformers)
        self.assertEqual(len(series), 2, 'Mismatched size {} != 2'.format(len(series)))
        series.append(FakeTransformer('fake3'))
        self.assertEqual(len(series), 3, 'Mismatched size {} != 3'.format(len(series)))

        del transformers



# -*- coding: utf-8 -*-
import logging

import numpy

from pipesnake.base import Transformer

__all__ = [
    'LSTMPacker',
]


class LSTMPacker(Transformer):
    """Pack rows in order to be used as input for LSTM networks.

    Args:
        :param sequence_len: the size of the sequences to be created (default: `2`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> packer = LSTMPacker(sequence_len=10)
    >>> x_new, y_new = packer.fit_transform(x, y)
    """

    def __init__(self, sequence_len=2, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.sequence_len = sequence_len

    def fit_x(self, x):
        assert type(x) == numpy.ndarray, 'x needs to be a numpy array, got: {}'.format(type(x))
        return self

    def fit_y(self, y):
        assert type(y) == numpy.ndarray, 'y needs to be a numpy array, got: {}'.format(type(y))
        return self

    def transform_x(self, x):
        data = []
        for i in range(x.shape[0] - self.sequence_len):
            data.append(x[i:i + self.sequence_len])
        return numpy.asarray(data)

    def transform_y(self, y):
        return y[self.sequence_len:]

    def inverse_transform_x(self, x):
        self.logging('LSTMPacker on x is not invertible', level=logging.WARNING)
        return x

    def inverse_transform_y(self, y):
        self.logging('LSTMPacker on x is not invertible', level=logging.WARNING)
        return y



# -*- coding: utf-8 -*-

from pipesnake.transformers.combiner import Roller

__all__ = [
    'ToReturn',
]


class ToReturn(Roller):
    """Convert columns to `financial return`: r_t = (x_t - x_{t-1}) / x_{t-1}.

    Note: resulting columns may contain inf and nan

    Args:
        :param x_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param y_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> returner = ToReturn(x_cols='all', y_cols='all')
    >>> x_new, y_new = returner.fit_transform(x, y)
    """

    @staticmethod
    def func(x):
        return (x[1] - x[0]) / float(x[0])

    def __init__(self, x_cols=[], y_cols=[], sklearn_output=False, name=None):
        Roller.__init__(self, x_cols=x_cols, y_cols=y_cols, roll_func=ToReturn.func, inv_roll_func=None,
                        window=2, sklearn_output=sklearn_output, name=name)



# -*- coding: utf-8 -*-

import logging

import networkx

from pipesnake.base.abstract import Pipe
from pipesnake.base.abstract import Transformer
from pipesnake.pipe.parallel import ParallelPipe
from pipesnake.pipe.series import SeriesPipe

__all__ = [
    'to_graph',
    'draw_graph_on_file',
]


def _get_edges(item, heads=[], edges=[]):
    """Computes all the edges (connections) between transformers.

    This is a bit convoluted, but it does the job

    :param item: a :class:`Pipe` inherited object
    :param heads: used for recursion
    :param relations: used for recursion
    :return: last head, and the list of edges
    """
    if isinstance(item, ParallelPipe):
        new_heads = []
        for sub_item in item.transformers:
            current_head, _ = _get_edges(sub_item, heads, edges)
            new_heads.extend(current_head)
        return new_heads, _
    elif isinstance(item, SeriesPipe):
        current_head = heads
        for sub_item in item.transformers:
            current_head, _ = _get_edges(sub_item, current_head, edges)
        return current_head, _
    elif isinstance(item, Transformer):
        for head in heads:
            edges.append((head, item.name))
    else:
        raise Exception('Unexpected object type: {}'.format(type(item)))
    return [item.name], edges


def _get_nodes(item, nodes=[]):
    """Get the list of nodes (aka list of :class:`Transformer`) and their params

    :param item: a :class:`Pipe` inherited object
    :param nodes: used for recursion
    :return: the list of nodes names and their params
    """
    if isinstance(item, ParallelPipe) or isinstance(item, SeriesPipe):
        for sub_item in item.transformers:
            _get_nodes(sub_item, nodes)
    elif isinstance(item, Transformer):
        nodes.append((item.name, item.get_params()))
    else:
        raise Exception('Unexpected object type: {}'.format(type(item)))
    return nodes


def to_graph(pipe):
    """Create a NetworkX representation of a pipe

    :param pipe: a pipesnake pipe
    :return: a NetworkX graph
    """
    assert issubclass(type(pipe), Pipe), 'Mismatched type: expected Pipe got {0}'.format(type(pipe))

    graph = networkx.Graph()

    for node in _get_nodes(pipe, nodes=[]):
        graph.add_node(node[0], params=node[1])

    _, edges = _get_edges(pipe, heads=[], edges=[])
    for _from, _to in edges:
        graph.add_edge(_from, _to)

    return graph


def draw_graph_on_file(graph, filename='figure.png'):
    """Save a png file with the pipe graph

    :param graph: a NetworkX graph
    :param filename: a filename for the image

    .. todo:: It would be nice to have a nice layout for this
    """
    try:
        from matplotlib import pyplot
    except Exception as e:
        logging.error('matplotlib is not installed')
        return
    pos = networkx.spring_layout(graph)
    networkx.draw(graph, pos=pos, with_labels=True, node_color='skyblue', node_shape='.', alpha=0.65)
    pyplot.savefig(filename)



# -*- coding: utf-8 -*-

from pipesnake.base import Transformer


class FakeTransformer(Transformer):
    def __init__(self, name):
        super(FakeTransformer, self).__init__(name=name)

    def fit(self, x, y):
        pass

    def transform(self, x, y):
        pass

    def inverse_transform(self, x, y):
        pass

    def get_params(self):
        pass

    def set_params(self, a, b):
        pass



# -*- coding: utf-8 -*-

"""
The :mod:`pipesnake.utils` module implements few utils functions, decorators and more.
"""

import re


def to_snake(name):
    """Convert :param name: to snake case"""
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



# -*- coding: utf-8 -*-

import logging
import math

from pipesnake.base import Transformer

__all__ = [
    'ToNumpy',
    'ColumnRenamer',
    'Copycat',
]


class ToNumpy(Transformer):
    """Convert `x` and `y` to a particular numpy type.

    The common case is that once you have your data pipeline you need to make it
    compatible with numpy so that scikit-learn can use it without any issue.
    Typically this transformer is the last one in the pipelines so that the output
    is ready to feed a scikit-learn pipeline.

    Note: this will not `y.ravel()` as far in pipesnake also target can be a matrix

    Args:
        :param as_type: is the final type you want to convert the dataframes (default: `float`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> astype = ToNumpy(as_type=int)
    >>> x_new, y_new = astype.fit_transform(x, y)
    """

    def __init__(self, as_type=float, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.as_type = as_type

    def fit_x(self, x):
        return self

    def fit_y(self, y):
        return self

    def _transform(self, data):
        _data = data.copy()
        return _data.astype(self.as_type).as_matrix()

    def transform_x(self, x):
        return self._transform(x)

    def transform_y(self, y):
        return self._transform(y)

    def inverse_transform_x(self, x):
        self.logging('type casting on x is not invertible as transformation', level=logging.WARNING)
        return x

    def inverse_transform_y(self, y):
        self.logging('type casting on x is not invertible as transformation', level=logging.WARNING)
        return y


class ColumnRenamer(Transformer):
    """Rename `x` and `y` columns.

    Args:
        :param prefix: prefix for column names (default: `''`)
        :param postfix: postfix for column names (default: `''`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> renamer = ColumnRenamer()
    >>> x_new, y_new = renamer.fit_transform(x, y)
    """

    def __init__(self, prefix='', postfix='', sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.prefix = prefix
        self.postfix = postfix
        # this is need to inverse mapping
        self._original_column_names = {'x': {}, 'y': {}}
        self._new_column_names = {'x': {}, 'y': {}}

    def _fit(self, data, col_name):
        original_names = data.columns.values.tolist()
        dim = len(original_names)
        pad = int(math.ceil(math.log10(float(dim))))
        base_name = self.prefix + col_name + '{}' + self.postfix
        new_names = [base_name.format(str(i).zfill(pad)) for i in range(dim)]
        return original_names, new_names

    def fit_x(self, x):
        self._original_column_names['x'], self._new_column_names['x'] = self._fit(x, 'x_')
        return self

    def fit_y(self, y):
        self._original_column_names['y'], self._new_column_names['y'] = self._fit(y, 'y_')
        return self

    def transform_x(self, x):
        _x = x.copy()
        _x.columns = self._new_column_names['x']
        self.logging('x new column names: {}'.format(self._new_column_names['x']), level=logging.DEBUG)
        return _x

    def transform_y(self, y):
        _y = y.copy()
        _y.columns = self._new_column_names['y']
        self.logging('y new column names: {}'.format(self._new_column_names['y']), level=logging.DEBUG)
        return _y

    def inverse_transform_x(self, x):
        _x = x.copy()
        _x.columns = self._original_column_names['x']
        self.logging('x original column names: {}'.format(self._original_column_names['x']), level=logging.DEBUG)
        return _x

    def inverse_transform_y(self, y):
        _y = y.copy()
        _y.columns = self._original_column_names['y']
        self.logging('y original column names: {}'.format(self._original_column_names['y']), level=logging.DEBUG)
        return _y


class Copycat(Transformer):
    """Copy the datasets forward.

    This is typically used to replicate dataset inside :class:`ParallelPipe` to augment features and/or targets.

    Args:
        :param on_x: apply on `x` (default: `True`)
        :param on_y: apply on `y` (default: `True`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`
    """

    def __init__(self, on_x=True, on_y=True, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.on_x = on_x
        self.on_y = on_y

    def fit_x(self, x):
        return self

    def fit_y(self, y):
        return self

    def transform_x(self, x):
        if not self.on_x:
            return None
        return x.copy()

    def transform_y(self, y):
        if not self.on_y:
            return None
        return y.copy()

    def inverse_transform_x(self, x):
        self.logging('copying x is not invertible', level=logging.WARNING)
        return x

    def inverse_transform_y(self, y):
        self.logging('copying x is not invertible', level=logging.WARNING)
        return y



# -*- coding: utf-8 -*-

import logging

import numpy

from pipesnake.base import Transformer
from pipesnake.base.utils import _check_cols

__all__ = [
    'MinMaxScaler',
    'StdScaler',
    'MadScaler',
    'UnitLenghtScaler',
]


class MinMaxScaler(Transformer):
    """Min max scaler.

    This will scale the data (by columns) to fit [min; max] interval.

    Args:
        :param x_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param y_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param min_value: min value after scaling (default: `1.0`)
        :param max_value: max value after scaling (default: `1.0`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> scaler = MinMaxScaler(x_cols=['x_1'], y_cols=['y_1'])
    >>> x_new, y_new = scaler.fit_transform(x, y)
    """

    def __init__(self, x_cols=[], y_cols=[], min_value=-1.0, max_value=1.0, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.x_cols = x_cols
        self.y_cols = y_cols
        self.min_value = min_value
        self.max_value = max_value
        self._d = self.max_value - self.min_value
        # the inverse mapping is needed to invert the transformation
        self._inverse_map = {'x': None, 'y': None}

    def fit_x(self, x):
        self.x_cols = _check_cols(x, self.x_cols, self.logging)
        self.logging('x_cols: {}'.format(self.x_cols), level=logging.DEBUG)
        return self

    def fit_y(self, y):
        self.y_cols = _check_cols(y, self.y_cols, self.logging)
        self.logging('y_cols: {}'.format(self.y_cols), level=logging.DEBUG)
        return self

    def transform_x(self, x):
        if len(self.x_cols) == 0:
            return x
        _x = x.copy()
        self._inverse_map['x'] = (_x[self.x_cols].min(), _x[self.x_cols].max())
        _x[self.x_cols] = (_x[self.x_cols] - _x[self.x_cols].min()) / (
            _x[self.x_cols].max() - _x[self.x_cols].min()) * self._d + self.min_value
        return _x

    def transform_y(self, y):
        if len(self.y_cols) == 0:
            return y
        _y = y.copy()
        self._inverse_map['y'] = (_y[self.y_cols].min(), _y[self.y_cols].max())
        _y[self.y_cols] = (_y[self.y_cols] - _y[self.y_cols].min()) / (
            _y[self.y_cols].max() - _y[self.y_cols].min()) * self._d + self.min_value
        return _y

    def inverse_transform_x(self, x):
        if len(self.x_cols) == 0:
            return x
        _x = x.copy()
        _min, _max = self._inverse_map['x']
        _x[self.x_cols] = (_x[self.x_cols] - self.min_value) / self._d * (_max - _min) + _min
        return _x

    def inverse_transform_y(self, y):
        if len(self._y_cols) == 0:
            return y
        _y = y.copy()
        _min, _max = self._inverse_map['y']
        _y[self.y_cols] = (_y[self.y_cols] - self.min_value) / self._d * (_max - _min) + _min
        return _y


class StdScaler(Transformer):
    """Standard deviation scaler.

    This will scale the data (by columns) following x' = ( x - mean(X) ) / ( k * std(X) ).

    Args:
        :param x_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param y_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param k: is a scaling factor for the standard deviation (default: `3.0`)
        :param skipna: exclude NA/null values when computing the result (default: `True`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> scaler = StdScaler(x_cols=['x_1'], y_cols=['y_1'])
    >>> x_new, y_new = scaler.fit_transform(x, y)
    """

    def __init__(self, x_cols=[], y_cols=[], k=3.0, skipna=True, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.x_cols = x_cols
        self.y_cols = y_cols
        self.k = k
        self.skipna = skipna
        # the inverse mapping is needed to invert the transformation
        self._inverse_map = {'x': {}, 'y': {}}

    def fit_x(self, x):
        self.x_cols = _check_cols(x, self.x_cols, self.logging)
        self.logging('x_cols: {}'.format(self.x_cols), level=logging.DEBUG)
        return self

    def fit_y(self, y):
        self.y_cols = _check_cols(y, self.y_cols, self.logging)
        self.logging('y_cols: {}'.format(self.y_cols), level=logging.DEBUG)
        return self

    def transform_x(self, x):
        if len(self.x_cols) == 0:
            return x
        _x = x.copy()
        _mean = _x[self.x_cols].mean(skipna=self.skipna)
        _std = _x[self.x_cols].std(skipna=self.skipna)
        self._inverse_map['x'] = (_mean, _std)
        _x[self.x_cols] = (_x[self.x_cols] - _mean) / (self.k * _std)
        return _x

    def transform_y(self, y):
        if len(self.y_cols) == 0:
            return y
        _y = y.copy()
        _mean = _y[self.y_cols].mean(skipna=self.skipna)
        _std = _y[self.y_cols].std(skipna=self.skipna)
        self._inverse_map['y'] = (_mean, _std)
        _y[self.y_cols] = (_y[self.y_cols] - _mean) / (self.k * _std)
        return _y

    def inverse_transform_x(self, x):
        if len(self.x_cols) == 0:
            return x
        _x = x.copy()
        _mean, _std = self._inverse_map['x']
        _x[self.x_cols] = _x[self.x_cols] * self.k * _std + _mean
        return _x

    def inverse_transform_y(self, y):
        if len(self.y_cols) == 0:
            return y
        _y = y.copy()
        _mean, _std = self._inverse_map['y']
        _y[self.y_cols] = _y[self.y_cols] * self.k * _std + _mean
        return _y


class MadScaler(Transformer):
    """Median absolute deviation scaler.

    This will scale the data (by columns) following x' = ( x - median(X) ) / ( k * mad(X) ).

    Args:
        :param x_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param y_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param k: is a scaling factor for the standard deviation (default: `3.0`)
        :param skipna: exclude NA/null values when computing the result (default: `True`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> scaler = MadScaler(x_cols=['x_1'], y_cols=['y_1'])
    >>> x_new, y_new = scaler.fit_transform(x, y)
    """

    def __init__(self, x_cols=[], y_cols=[], k=3.0, skipna=True, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.x_cols = x_cols
        self.y_cols = y_cols
        self.k = k
        self.skipna = skipna
        # the inverse mapping is needed to invert the transformation
        self._inverse_map = {'x': {}, 'y': {}}

    def fit_x(self, x):
        self.x_cols = _check_cols(x, self.x_cols, self.logging)
        self.logging('x_cols: {}'.format(self.x_cols), level=logging.DEBUG)
        return self

    def fit_y(self, y):
        self.y_cols = _check_cols(y, self.y_cols, self.logging)
        self.logging('y_cols: {}'.format(self.y_cols), level=logging.DEBUG)
        return self

    def transform_x(self, x):
        if len(self.x_cols) == 0:
            return x
        _x = x.copy()
        _median = _x[self.x_cols].median(skipna=self.skipna)
        _mad = _x[self.x_cols].mad(skipna=self.skipna)
        self._inverse_map['x'] = (_median, _mad)
        _x[self.x_cols] = (_x[self.x_cols] - _median) / (self.k * _mad)
        return _x

    def transform_y(self, y):
        if len(self.y_cols) == 0:
            return y
        _y = y.copy()
        _median = _y[self.y_cols].median(skipna=self.skipna)
        _mad = _y[self.y_cols].mad(skipna=self.skipna)
        self._inverse_map['y'] = (_median, _mad)
        _y[self.y_cols] = (_y[self.y_cols] - _median) / (self.k * _mad)
        return _y

    def inverse_transform_x(self, x):
        if len(self.x_cols) == 0:
            return x
        _x = x.copy()
        _median, _mad = self._inverse_map['x']
        _x[self.x_cols] = _x[self.x_cols] * self.k * _mad + _median
        return _x

    def inverse_transform_y(self, y):
        if len(self.y_cols) == 0:
            return y
        _y = y.copy()
        _median, _mad = self._inverse_map['y']
        _y[self.y_cols] = _y[self.y_cols] * self.k * _mad + _median
        return _y


class UnitLenghtScaler(Transformer):
    """Scale the feature vector to have norm 1.0.

    This will scale the data (by row) following X' = X / norm(X).

    Args:
        :param x_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param y_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param invertible: if true collect additional data make the transformer invertible. Note that
                           this requires to store one value for each row in `x` and `y` (default: `True`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this transformer

    Examples:

    >>> scaler = UnitLenghtScaler()
    >>> x_new, y_new = scaler.fit_transform(x, y)
    """

    def __init__(self, x_cols='all', y_cols=[], invertible=True, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.x_cols = x_cols
        self.y_cols = y_cols
        self.invertible = invertible
        # the inverse mapping is needed to invert the transformation
        self._inverse_map = {'x': None, 'y': None}

    def fit_x(self, x):
        self.x_cols = _check_cols(x, self.x_cols, self.logging)
        self.logging('x_cols: {}'.format(self.x_cols), level=logging.DEBUG)
        return self

    def fit_y(self, y):
        self.y_cols = _check_cols(y, self.y_cols, self.logging)
        self.logging('y_cols: {}'.format(self.y_cols), level=logging.DEBUG)
        return self

    def transform_x(self, x):
        if len(self.x_cols) == 0:
            return x
        _x = x.copy()
        n = numpy.sqrt(numpy.square(_x[self.x_cols]).sum(axis=1))
        if self.invertible:
            self._inverse_map['x'] = n
        _x[self.x_cols] = _x[self.x_cols].div(n, axis=0)
        return _x

    def transform_y(self, y):
        if len(self.y_cols) == 0:
            return y
        _y = y.copy()
        n = numpy.sqrt(numpy.square(_y[self.y_cols]).sum(axis=1))
        if self.invertible:
            self._inverse_map['y'] = n
        _y[self.y_cols] = _y[self.y_cols].div(n, axis=0)
        return _y

    def inverse_transform_x(self, x):
        if len(self.x_cols) == 0:
            return x
        if not self.invertible:
            self.logging('the transformer has not been set to be invertible, you should set invertible=True',
                         level=logging.WARNING)
            return x
        _x = x.copy()
        _x[self.x_cols] = _x[self.x_cols].mul(self._inverse_map['x'], axis=0)
        return _x

    def inverse_transform_y(self, y):
        if len(self.y_cols) == 0:
            return y
        if not self.invertible:
            self.logging('the transformer has not been set to be invertible, you should set invertible=True',
                         level=logging.WARNING)
            return y
        _y = y.copy()
        _y[self.y_cols] = _y[self.y_cols].mul(self._inverse_map['y'], axis=0)
        return _y



# -*- coding: utf-8 -*-

import logging

import pandas
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neighbors import KNeighborsRegressor

from pipesnake.base import Transformer
from pipesnake.base.utils import _check_cols
from pipesnake.base.utils import _is_categorical_cols
from pipesnake.base.utils import _shape
from pipesnake.transformers.converter import Category2Number

__all__ = [
    'ReplaceImputer',
    'KnnImputer',
]


class ReplaceImputer(Transformer):
    """Impute NaNs replacing them.

    Replace NaNs using different strategies. Note that categorical columns will not be imputed see :class:`Category2Number`

    Args:
        :param x_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param y_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param value: the value to be used in the imputation, it can be a number or `'mean'` or `'median'`, in the last
                      two cases the value is the mean or the median of the given column
        :param method:  {`backfill`, `bfill`, `pad`, `ffill`, `None`}, (default: `None`). Method to use for filling holes
                         in reindexed Series pad / ffill: propagate last valid observation forward to next valid
                         backfill / bfill: use NEXT valid observation to fill gap. More details: https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.fillna.html
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> imputer = ReplaceImputer()
    >>> x_new, y_new = imputer.fit_transform(x, y)
    """

    def __init__(self, x_cols=[], y_cols=[], value='median', method=None, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.x_cols = x_cols
        self.y_cols = y_cols
        self.value = value
        self.method = method

    def fit_x(self, x):
        self.x_cols = _check_cols(x, self.x_cols, self.logging)
        self.logging('x_cols: {}'.format(self.x_cols), level=logging.DEBUG)
        return self

    def fit_y(self, y):
        self.y_cols = _check_cols(y, self.y_cols, self.logging)
        self.logging('y_cols: {}'.format(self.y_cols), level=logging.DEBUG)
        return self

    def _transform(self, data, cols):
        if len(cols) == 0:
            return data
        _data = data.copy()
        value = {c: self.value for c in cols}
        if type(self.value) == str:
            if self.value == 'mean':
                value = _data[cols].mean().to_dict()
            if self.value == 'median':
                value = _data[cols].median().to_dict()
        _data.fillna(value=value, method=self.method, inplace=True)
        return _data

    def transform_x(self, x):
        return self._transform(x, self.x_cols)

    def transform_y(self, y):
        return self._transform(y, self.y_cols)

    def inverse_transform_x(self, x):
        self.logging('removing nan values in x is not invertible as transformation', level=logging.WARNING)
        return x

    def inverse_transform_y(self, y):
        self.logging('removing nan values in y is not invertible as transformation', level=logging.WARNING)
        return y


class KnnImputer(Transformer):
    """Impute NaNs using K-nearest neighbors.

    Replace NaNs using different Knn approach

    Args:
        :param x_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param y_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param k: k value for Knn algorithm
        :param x_is_categorical: a list of boolean indicating categorical columns, if `None` categorical will be autodetect
                                 using the type of the column, all non numerical columns will be treated as categorical (default: `None`)
        :param y_is_categorical: a list of boolean indicating categorical columns, if `None` categorical will be autodetect
                                 using the type of the column, all non numerical columns will be treated as categorical (default: `None`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> imputer = KnnImputer()
    >>> x_new, y_new = imputer.fit_transform(x, y)
    """

    def __init__(self, x_cols=[], y_cols=[], k=3, x_is_categorical=None, y_is_categorical=None, sklearn_output=False,
                 name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.x_cols = x_cols
        self.y_cols = y_cols
        self.k = k
        self.x_is_categorical = x_is_categorical
        self.y_is_categorical = y_is_categorical
        self._cat2num = {'x': None, 'y': None}  # this is needed to deal with potential categorical columns
        self._knn_models = {'x': None, 'y': None}

    def _fit(self, data, cols, is_categorical, cat2num):
        if len(cols) == 0:
            return data
        assert _shape(data)[1] >= 2, 'at least 2 columns are needed'
        _data = data.copy()

        nan_ids = pandas.isnull(data).any(axis=1).nonzero()[0].tolist()
        _data.drop(nan_ids, inplace=True)
        self.logging('current shape: {}'.format(_shape(_data)))
        assert _shape(_data)[0] >= self.k, 'not enough rows: {} removed {} rows'.format(_shape(_data)[0], len(nan_ids))

        _data, _ = cat2num.fit_transform(_data, None)

        models = {}
        for col, cat in zip(cols, is_categorical):
            if cat:
                models[col] = KNeighborsClassifier(n_neighbors=self.k)
            else:
                models[col] = KNeighborsRegressor(n_neighbors=self.k)
            x_train = _data[list(set(_data.columns.tolist()) - set([col]))]
            y_train = _data[col]
            self.logging(
                'fitting imputing_model for column: {} x_train.shape={} y_train.shape={}'.format(col, _shape(x_train),
                                                                                                 _shape(y_train)),
                                                                                                 level=logging.DEBUG)
            models[col].fit(x_train, y_train)
        return models

    def fit_x(self, x):
        self.x_cols = _check_cols(x, self.x_cols, self.logging)
        self.logging('x_cols: {}'.format(self.x_cols), level=logging.DEBUG)
        if not self.x_is_categorical or len(self.x_is_categorical) != len(self.x_cols):
            self.x_is_categorical = _is_categorical_cols(x, self.x_cols)
        categorical_cols = [self.x_cols[i] for i, v in enumerate(self.x_is_categorical) if v]
        self._cat2num['x'] = Category2Number(x_cols=categorical_cols, skipna=True)
        self._knn_models['x'] = self._fit(x, self.x_cols, self.x_is_categorical, self._cat2num['x'])
        return self

    def fit_y(self, y):
        self.y_cols = _check_cols(y, self.y_cols, self.logging)
        self.logging('y_cols: {}'.format(self.y_cols), level=logging.DEBUG)
        if not self.y_is_categorical or len(self.y_is_categorical) != len(self.y_cols):
            self.y_is_categorical = _is_categorical_cols(y, self.y_cols)
        categorical_cols = [self.y_cols[i] for i, v in enumerate(self.y_is_categorical) if v]
        self._cat2num['y'] = Category2Number(x_cols=categorical_cols, skipna=True)  # yes, it is x_col
        self._knn_models['y'] = self._fit(y, self.y_cols, self.y_is_categorical, self._cat2num['y'])
        return self

    def _transform(self, data, cols, cat2num, models):
        if len(cols) == 0:
            return data
        _data = data.copy()
        _data, _ = cat2num.transform(_data, None)

        # compute imputation values
        impute_values = {}
        for col in _data.columns.values.tolist():
            impute_values[col] = _data[col].mode().iloc[0]

        for col in cols:
            nan_ids = pandas.isnull(_data[col]).nonzero()[0].tolist()
            if len(nan_ids) == 0:
                continue
            x_pred = _data.loc[nan_ids].copy()
            x_pred.drop([col], axis=1, inplace=True)
            x_pred.fillna(value=impute_values, inplace=True)  # temporary imputation for other missing values
            self.logging('predicting model for column: {}'.format(col), level=logging.DEBUG)
            _data.loc[nan_ids, col] = models[col].predict(x_pred)

        _data, _ = cat2num.inverse_transform(_data, None)
        return _data

    def transform_x(self, x):
        return self._transform(x, self.x_cols, self._cat2num['x'], self._knn_models['x'])

    def transform_y(self, y):
        return self._transform(y, self.y_cols, self._cat2num['y'], self._knn_models['y'])

    def inverse_transform_x(self, x):
        self.logging('removing nan cols in x is not invertible as transformation', level=logging.WARNING)
        return x

    def inverse_transform_y(self, y):
        self.logging('removing nan cols y is not invertible as transformation', level=logging.WARNING)
        return y



# -*- coding: utf-8 -*-
"""Metadata for pipesnake"""

# The package name, which is also the "UNIX name" for the project.
package = 'pipesnake'
project = 'Pipesnake'
project_no_spaces = project.replace(' ', '')
version = '0.1'
description = 'Feature extractor and data processing pipelines for Pandas inspired by Scikit-Learn.'
authors = ['Pierluigi Failla']
authors_string = ', '.join(authors)
emails = ['{name}.{surname}@gmail.com']
license = 'MIT'
copyright = '2018 ' + authors_string
url = 'https://github.com/pierluigi-failla/pipesnake'



# -*- coding: utf-8 -*-

import unittest

from pipesnake.transformers.dropper import DropDuplicates
from pipesnake.transformers.dropper import DropNanRows


class TestDropper(unittest.TestCase):
    def test_drop_duplicates(self):
        from _data import x
        from _data import y

        dropper = DropDuplicates()
        x_new, y_new = dropper.fit_transform(x, y)
        self.assertEqual(x_new.shape[1], 1, 'x mismatch sizes: {} != 1'.format(x_new.shape[1]))
        self.assertEqual(y_new.shape[1], 2, 'y mismatch sizes: {} != 1'.format(y_new.shape[1]))

        del x, y, x_new, y_new

    def test_drop_nan_rows(self):
        from _data import x_nan
        from _data import y_nan

        dropper = DropNanRows()
        x_new, y_new = dropper.fit_transform(x_nan, y_nan)
        self.assertEqual(x_new.shape[0], 4, 'x mismatch sizes: {} != 1'.format(x_new.shape[0]))
        self.assertEqual(y_new.shape[0], 4, 'y mismatch sizes: {} != 1'.format(x_new.shape[0]))

        del x_nan, y_nan, x_new, y_new



# -*- coding: utf-8 -*-

import logging

import pandas

from pipesnake.base import Transformer
from pipesnake.base.utils import _shape

__all__ = [
    'DropDuplicates',
    'DropNanCols',
    'DropNanRows',
]


class DropDuplicates(Transformer):
    """Drop duplicated rows and/or cols.

    Args:
        :param on_rows: remove duplicated rows (default: `True`)
        :param on_cols: remove duplicated columns (default: `True`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> dropper = DropDuplicates()
    >>> x_new, y_new = dropper.fit_transform(x, y)
    """

    def __init__(self, on_rows=True, on_cols=True, on_x=True, on_y=False, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.on_rows = on_rows
        self.on_cols = on_cols
        self.on_x = on_x
        self.on_y = on_y
        self._drop_indices_row = []

    def fit_x(self, x):
        if not self.on_x:
            return self
        if self.on_rows:
            self._drop_indices_row.extend(x.index[x.duplicated()].tolist())
            self._drop_indices_row = list(set(self._drop_indices_row))
        return self

    def fit_y(self, y):
        if not self.on_y:
            return self
        if self.on_rows:
            self._drop_indices_row.extend(y.index[y.duplicated()].tolist())
            self._drop_indices_row = list(set(self._drop_indices_row))
        return self

    def _transform(self, data, on_cols):
        _data = data.copy()
        if len(self._drop_indices_row) > 0:
            _data.drop(self._drop_indices_row, inplace=True)
        if self.on_cols and on_cols:
            _data = _data.T.drop_duplicates()
            _data = _data.T
        self.logging('  shape: {}'.format(_shape(_data)), level=logging.DEBUG)
        return _data

    def transform_x(self, x):
        self.logging('x shape: {}'.format(_shape(x)), level=logging.DEBUG)
        return self._transform(x, self.on_x)

    def transform_y(self, y):
        self.logging('y shape: {}'.format(_shape(y)), level=logging.DEBUG)
        return self._transform(y, self.on_y)

    def inverse_transform_x(self, x):
        self.logging('removing x duplicates is not invertible as transformation', level=logging.WARNING)
        return x

    def inverse_transform_y(self, y):
        self.logging('removing y duplicates is not invertible as transformation', level=logging.WARNING)
        return y


class DropNanCols(Transformer):
    """Drop cols with nans.

    Args:
        :param how: 'any' if any NA values are present, 'all' if all values are NA (default: `'any'`). More details: https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.dropna.html
        :param on_x: apply on `x` (default: `True`)
        :param on_y: apply on `y` (default: `True`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> dropper = DropNanCols()
    >>> x_new, y_new = dropper.fit_transform(x, y)
    """

    def __init__(self, how='any', on_x=True, on_y=True, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.how = how
        self.on_x = on_x
        self.on_y = on_y

    def fit_x(self, x):
        return self

    def fit_y(self, y):
        return self

    def _transform(self, data):
        _data = data.copy()
        _data.dropna(axis=1, how=self.how, inplace=True)
        self.logging('  shape: {}'.format(_shape(_data)), level=logging.DEBUG)
        return _data

    def transform_x(self, x):
        if not self.on_x:
            return x
        self.logging('x shape: {}'.format(_shape(x)), level=logging.DEBUG)
        return self._transform(x)

    def transform_y(self, y):
        if not self.on_y:
            return y
        self.logging('y shape: {}'.format(_shape(y)), level=logging.DEBUG)
        return self._transform(y)

    def inverse_transform_x(self, x):
        self.logging('removing nan cols in x is not invertible as transformation', level=logging.WARNING)
        return x

    def inverse_transform_y(self, y):
        self.logging('removing nan cols y is not invertible as transformation', level=logging.WARNING)
        return y


class DropNanRows(Transformer):
    """Drop rows with `any` or `all` nans.

    Args:
        :param how: 'any' if any NA values are present, 'all' if all values are NA (default: `'any'`) More details: https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.dropna.html
        :param on_x: apply on `x` (default: `True`)
        :param on_y: apply on `y` (default: `True`)
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> dropper = DropNanRows()
    >>> x_new, y_new = dropper.fit_transform(x, y)
    """

    def __init__(self, how='any', on_x=True, on_y=True, sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.how = how
        self.on_x = on_x
        self.on_y = on_y
        self._drop_indices = []

    def _fit(self, data, apply):
        if apply:
            if self.how == 'any':
                self._drop_indices.extend(pandas.isnull(data).any(axis=1).nonzero()[0].tolist())
            if self.how == 'all':
                self._drop_indices.extend(pandas.isnull(data).all(axis=1).nonzero()[0].tolist())
            self._drop_indices = list(set(self._drop_indices))  # remove duplicates

    def fit_x(self, x):
        self._fit(x, self.on_x)
        return self

    def fit_y(self, y):
        self._fit(y, self.on_y)
        return self

    def _transform(self, data):
        _data = data.copy()
        _data.drop(self._drop_indices, inplace=True)
        self.logging('  shape: {}'.format(_shape(_data)), level=logging.DEBUG)
        return _data

    def transform_x(self, x):
        self.logging('x shape: {}'.format(_shape(x)), level=logging.DEBUG)
        return self._transform(x)

    def transform_y(self, y):
        self.logging('y shape: {}'.format(_shape(y)), level=logging.DEBUG)
        return self._transform(y)

    def inverse_transform_x(self, x):
        self.logging('removing nan rows in x is not invertible as transformation', level=logging.WARNING)
        return x

    def inverse_transform_y(self, y):
        self.logging('removing nan rows y is not invertible as transformation', level=logging.WARNING)
        return y



# -*- coding: utf-8 -*-

import logging

from pipesnake.base import Transformer
from pipesnake.base.utils import _check_cols
from pipesnake.base.utils import _shape

__all__ = [
    'ColumnSelector',
]


class ColumnSelector(Transformer):
    """Select a given list of column names to keep

    Args:
        :param x_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param y_cols: a list of columns name or a list of indices; 'all' to use all columns; if [] no columns will be affected
        :param sklearn_output: if True produces outputs compatible with sklearn Pipeline (default: `False`)
        :param name: name for this :class:`Transformer`

    Examples:

    >>> selector = ColumnSelector(x_cols=['column_A', 7], y_cols=[0])
    >>> # This will return an x DataFrame with only 'column_A' and the 8th column, and a y DataFrame
    >>> # with just the first column
    """

    def __init__(self, x_cols=[], y_cols=[], sklearn_output=False, name=None):
        Transformer.__init__(self, sklearn_output=sklearn_output, name=name)
        self.x_cols = x_cols
        self.y_cols = y_cols

    def fit_x(self, x):
        """Create the list of columns to be kept for x.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Return:
            :param self:
        """
        self.x_cols = _check_cols(x, self.x_cols, self.logging)
        self.logging('x_cols: {}'.format(self.x_cols), level=logging.DEBUG)
        return self

    def fit_y(self, y):
        """Create the list of columns to be kept for y.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Return:
            :param self:
        """
        self.y_cols = _check_cols(y, self.y_cols, self.logging)
        self.logging('y_cols: {}'.format(self.y_cols), level=logging.DEBUG)
        return self

    def transform_x(self, x):
        """Transform x just dropping irrelevant columns.

        Args:
            :param x: a Pandas Dataframe of shape [n_samples, n_features] the dataset
        Returns:s
            :return x_new: the new transformed x
        """
        if len(self.x_cols) == 0:
            return x
        self.logging('x shape: {}'.format(_shape(x)), level=logging.DEBUG)
        x_new = x.copy()
        if len(self.x_cols) > 0:
            x_new.drop(list(set(x_new.columns.values.tolist()) - set(self.x_cols)), axis=1, inplace=True)
        self.logging(' shape: {}'.format(_shape(x_new)), level=logging.DEBUG)
        return x_new

    def transform_y(self, y):
        """Transform y just dropping irrelevant columns.

        Args:
            :param y: a Pandas Dataframe of shape [n_samples] the target
        Returns:
            :return y_new: the new transformed y
        """
        if len(self.y_cols) == 0:
            return y
        self.logging('y shape: {}'.format(_shape(y)), level=logging.DEBUG)
        y_new = y.copy()
        if len(self.y_cols) > 0:
            y_new.drop(list(set(y_new.columns.values.tolist()) - set(self.y_cols)), axis=1, inplace=True)
        self.logging('  shape: {}'.format(_shape(y_new)), level=logging.DEBUG)
        return y_new

    def inverse_transform_x(self, x):
        """Removing columns is not invertible as transformation.

        Returns:
            :return x: no inverse transform is applied
        """
        self.logging('removing x columns is not invertible as transformation', level=logging.WARNING)
        return x

    def inverse_transform_y(self, y):
        """Removing columns is not invertible as transformation.

        Returns:
            :return y: no inverse transform is applied
        """
        self.logging('removing y columns is not invertible as transformation', level=logging.WARNING)
        return y



# -*- coding: utf-8 -*-

import logging

import pandas


def _check_transformer_type(obj, logger):
    """ Check if a given object is :class:`Transformer`

    :param obj: the instance
    :return: True if it is :class:`Transformer` False otherwise
    """
    if not (hasattr(obj, 'fit') and callable(obj.fit)):
        logger('{} seems not be a Transformer'.format(type(obj)), level=logging.WARNING)
        return False
    if not (hasattr(obj, 'transform') and callable(obj.transform)):
        logger('{} seems not be a Transformer'.format(type(obj)), level=logging.WARNING)
        return False
    if not (hasattr(obj, 'fit_transform') and callable(obj.fit_transform)):
        logger('{} seems not be a Transformer'.format(type(obj)), level=logging.WARNING)
        return False
    if not (hasattr(obj, 'inverse_transform') and callable(obj.inverse_transform)):
        logger('{} seems not be a Transformer'.format(type(obj)), level=logging.WARNING)
        return False
    return True


def _check_input(df, logger):
    """Check input type and number or rows

    Args:
        :param df: a pandas DataFrame
    Return:
        :return: True if everything is ok
    """
    if '.DataFrame' in str(type(df)):
        if len(df) > 0:
            return True
        else:
            logger('input has not enough rows, got shape {}'.format(df.shape), level=logging.WARNING)
            return False
    else:
        logger('input is not a Pandas DataFrame, instead got {}'.format(type(df)), level=logging.WARNING)
        return False


def _check_cols(df, cols, logger):
    """Produce a correct list of columns for this DataFrame

    Args:
        :param df: a pandas DataFrame
        :param cols: a list of column names or a list of indices; 'all' to use all columns; if [] no columns will be affected
    Return:
        :return: a convalidated list of columns for this DataFrame
    Raise:
        :exception: unknown type of columns
    """

    if type(cols) == str and cols == 'all':
        return df.columns.values.tolist()
    elif type(cols) == list:
        _cols = []
        for c in cols:
            if type(c) == int:
                _cols.append(df.columns.values[c])
            else:
                _cols.append(c)
        return _cols
    else:
        logger('unknown type of columns: {} ({})'.format(type(cols), cols), level=logging.WARNING)
        raise Exception('unknown type of columns: {} ({})'.format(type(cols), cols))


def _is_categorical_cols(df, cols):
    """Detect if column type is a number or string

    Args:
        :param df: a pandas DataFrame
        :param cols: a list of column names
    Return:
        :return: a boolean list of same lenght as cols with True in position where the column is not numerical
    Raise:
        :exception: unknown type of columns
    """
    is_categorical = []
    for c in cols:
        is_categorical.append(not pandas.np.issubdtype(df[c].dtype, pandas.np.number))
    return is_categorical


def _shape(df):
    """ Return DataFrame shape even if is not a Pandas dataframe."""
    if type(df) == pandas.DataFrame or type(df) == pandas.Series:
        return df.shape
    try:
        shape = (len(df), len(df.columns))
    except Exception as e:
        logging.error(e)
        raise e
    return shape



# -*- coding: utf-8 -*-

import unittest

from pipesnake.pipe import ParallelPipe
from pipesnake.pipe import SeriesPipe
from pipesnake.transformers.selector import ColumnSelector


class TestPipe(unittest.TestCase):
    def test_aseries(self):
        from _data import x
        from _data import y

        transformers = [ColumnSelector(x_cols=['x_3', 'x_4', ]),
                        ColumnSelector(x_cols=['x_4', ], y_cols=['y_1', ]), ]
        pipe = SeriesPipe(transformers=transformers)
        x_new, y_new = pipe.fit_transform(x, y)
        self.assertEqual(len(pipe), len(transformers),
                         'Mismatched size {} != {}'.format(len(pipe), len(transformers)))
        self.assertEqual(len(x_new.columns), 1, 'Mismatched size {} != 1'.format(len(x_new.columns)))
        self.assertEqual(len(y_new.columns), 1, 'Mismatched size {} != 1'.format(len(y_new.columns)))

        del x, y, x_new, y_new

    def test_parallel(self):
        from _data import x
        from _data import y

        transformers = [ColumnSelector(x_cols=['x_1', 'x_2', 'x_3', 'x_4', ], y_cols=['y_1', ], name='1'),
                        ColumnSelector(x_cols=['x_3', ], name='2'), ]
        pipe = ParallelPipe(transformers=transformers)
        x_new, y_new = pipe.fit_transform(x, y)
        self.assertEqual(len(pipe), len(transformers),
                         'Mismatched size {} != {}'.format(len(pipe), len(transformers)))
        self.assertEqual(len(x_new.columns), 5, 'Mismatched size {} != 5'.format(len(x_new.columns)))
        self.assertEqual(len(y_new.columns), 3, 'Mismatched size {} != 3'.format(len(y_new.columns)))

        del x, y, x_new, y_new

    def test_both(self):
        from _data import x
        from _data import y

        pipe = SeriesPipe(transformers=[
            ParallelPipe(transformers=[
                ColumnSelector(x_cols=['x_1', 'x_2', 'x_3', 'x_4', ], y_cols=['y_1', ], name='1'),
                ColumnSelector(x_cols=['x_3', ], name='2'),
            ]),
            ColumnSelector(x_cols=['x_1_1', ], y_cols=['y_1_1', ], name='3'),
        ])
        x_new, y_new = pipe.fit_transform(x, y)
        self.assertEqual(len(x_new.columns), 1, 'Mismatched size {} != 1'.format(len(x_new.columns)))
        self.assertEqual(len(y_new.columns), 1, 'Mismatched size {} != 1'.format(len(y_new.columns)))

        del x, y, x_new, y_new
